"""Helper flax.nn.Modules for composing models

Todo:
    * Implement `WeightedParallel` with trainable weights
"""
from typing import List

import flax
import numpy as onp


class Sequential(flax.nn.Module):
    """Create a module from a sequential set of modules

    Notes:
        * Takes the output of the first model, passes in to the second, etc
    """

    def apply(self, x: onp.ndarray, learners: List[flax.nn.Model]):
        """
        Args:
            x (onp.ndarray): input data
            modules (List[flax.nn.module]): list of flax modules
            args (List[dict]): list of kwargs corresponding to the `modules`
                argument to initialize modules

        returns:
            onp.ndarray: result
        """
        result = x
        for learner in learners:
            result = learner(result)
        return result


class Parallel(flax.nn.Module):
    """Create a module from a sequential set of modules

    Notes:
        * Return a list of outputs from each model
    """

    def apply(self, x: onp.ndarray, learners: List[flax.nn.Model]):
        """
        Args:
            x (onp.ndarray): input data
            modules (List[flax.nn.module]): list of flax modules
            args (List[dict]): list of kwargs corresponding to the `modules`
                argument to initialize modules

        returns:
            List[onp.ndarray]: list of results
        """
        return [learner(x) for learner in learners]



"""timecast.series.unemployment"""
import os
from typing import Tuple

import numpy as onp

from timecast.series._core import generate_timeline


def generate(path=None) -> Tuple[onp.ndarray, onp.ndarray]:
    """
    Description: Monthly unemployment rate since 1948.

    References:
        * https://fred.stlouisfed.org/series/UNRATE
    """

    return generate_timeline(
        path
        or os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../data/unemployment.csv"),
        name="UNRATE",
    )



"timecast.series.random"
from typing import Tuple

import jax.numpy as jnp
import numpy as onp


def generate(
    n: int = 1000, loc: float = 0.0, scale: float = 1.0
) -> Tuple[onp.ndarray, onp.ndarray]:
    """
    Description: outputs a timeline randomly distributed i.i.d. from gaussian
    with mean `loc`, standard deviation `scale`
    """
    X = onp.random.normal(loc=loc, scale=scale, size=(n + 1))

    return jnp.asarray(X)[:-1].reshape(-1, 1), jnp.asarray(X)[1:].reshape(-1, 1)



"""Type and size check all series"""
import jax.numpy as jnp
import pytest

from timecast.series import arma
from timecast.series import crypto
from timecast.series import enso
from timecast.series import lds
from timecast.series import lstm
from timecast.series import random
from timecast.series import rnn
from timecast.series import sp500
from timecast.series import uci
from timecast.series import unemployment


@pytest.mark.parametrize(
    "module", [arma, crypto, enso, lds, lstm, random, rnn, sp500, uci, unemployment]
)
def test_series(module):
    """Test outputs, shapes of data time series"""
    X, y = getattr(module, "generate")()  # noqa: B009

    assert isinstance(X, jnp.ndarray)
    assert isinstance(y, jnp.ndarray)
    assert X.shape[0] == y.shape[0]



"""timecast.utils.random: testing"""
import jax.numpy as jnp

from timecast.utils import random


def test_random():
    """Test random singleton"""
    random.set_key()
    random.set_key(0)
    assert jnp.array_equal(jnp.array([0, 0]), random.get_key())
    expected = jnp.array([2718843009, 1272950319], dtype=jnp.uint32)
    assert jnp.array_equal(random.generate_key(), expected)
    expected = jnp.array([4146024105, 967050713], dtype=jnp.uint32)
    assert jnp.array_equal(random.get_key(), expected)



"""timecast.learners.PredictConstant: testing"""
import flax
import jax
import jax.numpy as jnp
import numpy as onp
import pytest

from timecast.learners import PredictConstant
from timecast.utils import random

shapes = [(4, 32), (10,), (10, 1), (1,), (1, 10)]


def create_predict_constant(shape, c):
    """Create PredictConstant model"""
    model_def = PredictConstant.partial(c=c)
    with flax.nn.stateful() as state:
        _, params = model_def.init_by_shape(jax.random.PRNGKey(0), [shape])
        model = flax.nn.Model(model_def, params)
    return model, state


@pytest.mark.parametrize("shape", shapes)
@pytest.mark.parametrize("c", [-1, 0, 1, 0.1])
def test_predict_constant_scalar(shape, c):
    """Test scalar constant"""
    model, state = create_predict_constant(shape, c)

    X = jax.random.uniform(random.generate_key(), shape=shape)
    with flax.nn.stateful(state) as state:
        ys = model(X)

    onp.testing.assert_array_almost_equal(jnp.ones_like(X) * c, ys)


@pytest.mark.parametrize("shape", shapes)
def test_predict_constant_vector(shape):
    """Test vector constant"""
    c = jax.random.uniform(random.generate_key(), shape=shape)
    model, state = create_predict_constant(shape, c)

    X = jax.random.uniform(random.generate_key(), shape=shape)
    with flax.nn.stateful(state) as state:
        ys = model(X)

    onp.testing.assert_array_almost_equal(c, ys)


@pytest.mark.parametrize("shape", shapes)
def test_predict_constant_exception(shape):
    """Test exceptional behavior"""
    shape = (dim + 1 for dim in shape)
    c = jax.random.uniform(random.generate_key(), shape=shape)
    with pytest.raises(ValueError):
        model, state = create_predict_constant(shape, c)



"""timecast.series.enso"""
import os
from typing import Tuple

import jax.numpy as jnp
import numpy as onp
import pandas as pd


def generate(input_signals=None, output_signals=None, path=None) -> Tuple[onp.ndarray, onp.ndarray]:
    """Generate ENSO

    Description: Collection of monthly values of control indices useful for predicting
    La Nina/El Nino. More specifically, the user can choose any of pna, ea,
    wa, wp, eu, soi, esoi, nino12, nino34, nino4, oni of nino34 (useful for
    La Nino/El Nino identification) to be used as input and/or output in
    the problem instance.
    """

    input_signals = input_signals or [
        "pna",
        "ea",
        "wa",
        "wp",
        "eu",
        "soi",
        "esoi",
        "nino12",
        "nino34",
        "nino4",
    ]
    output_signals = output_signals or ["oni"]
    data = pd.read_csv(
        path or os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../data/enso.csv")
    )
    signal_length = data.shape[0]

    # Get climatology
    clm = jnp.asarray(data.nino34).reshape(-1, 12).mean(axis=0)

    # Compute anomaly
    data["anm"] = (jnp.asarray(data.nino34).reshape(-1, 12) - clm).reshape(signal_length)

    # Get ONI
    data["oni"] = jnp.asarray(data.anm.rolling(3, min_periods=1).mean())
    data["month"] = jnp.arange(signal_length) % 12

    return (
        jnp.asarray(data[input_signals]),
        jnp.asarray(data[output_signals]).reshape(-1, 1),
    )



"""timecast.utils.statistics: testing"""
import jax
import numpy as onp
import pytest

from timecast.utils import random
from timecast.utils.statistics import OnlineStatistics


@pytest.mark.parametrize("n", [2, 10, 50, 100])
@pytest.mark.parametrize("j", [1, 10, 100])
@pytest.mark.parametrize("k", [1, 10, 100])
@pytest.mark.parametrize("func", ["sum", "mean", "std", "var", "observations", "zscore"])
def test_online_sum(n, j, k, func):
    """Test online statistics"""
    stats = OnlineStatistics(dim=k)
    X = jax.random.uniform(random.generate_key(), shape=(n, j * k))

    for i in X:
        stats.update(i.reshape(j, k))

    if func == "zscore":
        onp.testing.assert_array_almost_equal(
            stats.zscore(X[0, :].reshape(j, k)), (X[0, :].reshape(j, k) - stats.mean) / stats.std,
        )
    elif func != "observations":
        result = getattr(stats, func)
        onp.testing.assert_array_almost_equal(
            result, getattr(X.reshape(n * j, k), func)(axis=0).reshape(1, -1), decimal=2
        )
    else:
        assert n * j == stats.observations



"""timecast.experiment: testing"""
import jax
import numpy as onp
import pytest

from timecast import experiment
from timecast.utils import random


@pytest.mark.parametrize("shape", [(), (1,), (1, 2), (1, 2, 3)])
@pytest.mark.parametrize("num_args", [1, 2, 10])
def test_experiment(shape, num_args):
    """Test normal experiment behavior"""
    args = [
        (
            jax.random.uniform(random.generate_key(), shape=shape),
            jax.random.uniform(random.generate_key(), shape=shape),
        )
        for _ in range(num_args)
    ]

    @experiment("a,b", args)
    def dummy(a, b):
        """dummy"""
        return a + b

    results = dummy.run()
    print(results)
    for i in range(len(results)):
        onp.testing.assert_array_almost_equal(results[i], onp.sum(args[i], axis=0))


@pytest.mark.parametrize("times", [1, 2, 10])
def test_experiment_call(times):
    """Tests repeated decorator calls"""

    def dummy(a, b):
        """dummy"""
        return a + b

    for _ in range(times):
        dummy = experiment("a,b", [(1, 2)])(dummy)
        dummy._func(1, 2)
        assert isinstance(dummy, experiment)


@pytest.mark.parametrize("arglists", [[()], [(1, 2), ()], [(1,)]])
def test_experiment_inconsistent_arglist_len(arglists):
    """Tests inconsistent arglist length"""
    with pytest.raises(ValueError):

        def dummy(a, b):
            """dummy"""
            return a + b

        dummy(1, 2)
        experiment("a,b", arglists)(dummy)


def test_experiment_duplicated_argname():
    """Tests duplicated arguments"""
    with pytest.raises(ValueError):

        @experiment("a", [1])
        @experiment("a,b", [(1, 2)])
        def dummy(a, b):
            """dummy"""
            return a + b

        dummy._func(1, 2)
        dummy._validate()


def test_experiment_missing_argument():
    """Test missing arguments"""
    with pytest.raises(ValueError):

        @experiment("a", [1])
        def dummy(a, b):
            """dummy"""
            return a + b

        dummy._func(1, 2)
        dummy._validate()


def test_experiment_unused_arguments():
    """Testing unused arguments"""
    with pytest.raises(ValueError):

        @experiment("a,b,c", [(1, 2, 3)])
        def dummy(a, b):
            """dummy"""
            return a + b

        dummy._func(1, 2)
        dummy._validate()


def test_experiment_list_args():
    """Testing args as list"""

    @experiment(["a", "b"], [(1, 2)])
    def dummy(a, b):
        """dummmy"""
        return a + b

    dummy.run()


def test_experiment_list_atoms():
    """Testing atoms"""

    @experiment(["a"], [1])
    def dummy(a):
        """dummy"""
        return a

    assert 1 == dummy.run()[0]



"""timecast.utils.gram: testing"""
import jax
import jax.numpy as jnp
import numpy as onp
import pytest

from timecast.utils import random
from timecast.utils.gram import OnlineGram


dims = [1, 10, 20]


@pytest.mark.parametrize("X_dim", dims)
@pytest.mark.parametrize("Y_dim", dims)
@pytest.mark.parametrize("n", [1, 10, 20])
def test_gram_update(X_dim, Y_dim, n):
    """Test update"""
    X = jax.random.uniform(random.generate_key(), shape=(n, X_dim))
    Y = jax.random.uniform(random.generate_key(), shape=(n, Y_dim))

    gram = OnlineGram(X_dim, Y_dim, fit_intercept=False, normalize=False)
    gram.update(X, Y)

    onp.testing.assert_array_almost_equal(gram.matrix, X.T @ Y)


@pytest.mark.parametrize("X_dim", dims)
@pytest.mark.parametrize("Y_dim", dims)
@pytest.mark.parametrize("n", [1, 10, 20])
def test_gram_update_iterative(X_dim, Y_dim, n):
    """Test update iteratively"""
    X = jax.random.uniform(random.generate_key(), shape=(n, X_dim))
    Y = jax.random.uniform(random.generate_key(), shape=(n, Y_dim))

    gram = OnlineGram(X_dim, Y_dim, fit_intercept=False, normalize=False)

    for x, y in zip(X, Y):
        gram.update(x.reshape(1, -1), y.reshape(1, -1))

    onp.testing.assert_array_almost_equal(gram.matrix, X.T @ Y, decimal=3)


def test_gram_update_value_error():
    """Test update value error"""
    gram = OnlineGram(1, 1)
    with pytest.raises(ValueError):
        gram.update(
            jax.random.uniform(random.generate_key(), shape=(4, 1)),
            jax.random.uniform(random.generate_key(), shape=(1, 1)),
        )

    with pytest.raises(ValueError):
        gram.update(jax.random.uniform(random.generate_key(), shape=(4, 1)))

    gram = OnlineGram(1)
    with pytest.raises(ValueError):
        gram.update(onp.ones((1, 1)), onp.ones((1, 1)))


@pytest.mark.parametrize("X_dim", dims)
@pytest.mark.parametrize("Y_dim", dims)
@pytest.mark.parametrize("n", [2, 10, 20])
def test_gram_normalize(X_dim, Y_dim, n):
    """Test normalize"""
    X = jax.random.uniform(random.generate_key(), shape=(n, X_dim))
    Y = jax.random.uniform(random.generate_key(), shape=(n, Y_dim))

    gram = OnlineGram(X_dim, Y_dim, fit_intercept=False, normalize=True)
    gram.update(X, Y)

    X_norm = (X - X.mean(axis=0)) / X.std(axis=0)
    Y_norm = (Y - Y.mean(axis=0)) / Y.std(axis=0)
    onp.testing.assert_array_almost_equal(gram.matrix, X_norm.T @ Y_norm, decimal=1)


@pytest.mark.parametrize("X_dim", dims)
@pytest.mark.parametrize("n", [3, 10, 20])
@pytest.mark.parametrize("normalize", [True, False])
def test_gram_intercept_xtx(X_dim, n, normalize):
    """Test intercept on X.T @ X"""
    X = jax.random.uniform(random.generate_key(), shape=(n, X_dim))

    gram = OnlineGram(X_dim, fit_intercept=True, normalize=normalize)
    gram.update(X)

    if normalize:
        X = (X - X.mean(axis=0)) / X.std(axis=0)

    X = jnp.hstack((jnp.ones((n, 1)), X))

    onp.testing.assert_array_almost_equal(gram.matrix, X.T @ X, decimal=1)


@pytest.mark.parametrize("X_dim", dims)
@pytest.mark.parametrize("Y_dim", dims)
@pytest.mark.parametrize("n", [2, 10, 20])
@pytest.mark.parametrize("normalize", [True, False])
def test_gram_intercept_xty(X_dim, Y_dim, n, normalize):
    """Test intercept on X.T @ Y"""
    X = jax.random.uniform(random.generate_key(), shape=(n, X_dim))
    Y = jax.random.uniform(random.generate_key(), shape=(n, Y_dim))

    gram = OnlineGram(X_dim, Y_dim, fit_intercept=True, normalize=normalize)

    gram.update(X, Y)

    onp.testing.assert_array_almost_equal(gram.mean.squeeze(), X.mean(axis=0))
    onp.testing.assert_array_almost_equal(gram.std.squeeze(), X.std(axis=0))
    assert gram.observations == X.shape[0]

    if normalize:
        X = (X - X.mean(axis=0)) / X.std(axis=0)
        Y = (Y - Y.mean(axis=0)) / Y.std(axis=0)

    X = jnp.hstack((jnp.ones((n, 1)), X))

    onp.testing.assert_array_almost_equal(gram.matrix, X.T @ Y, decimal=1)



"""timecast.learners.AR: testing"""
import flax
import jax
import jax.numpy as jnp
import numpy as onp
import pytest
from sklearn.linear_model import Ridge

from timecast.learners import AR
from timecast.learners._ar import _compute_kernel_bias
from timecast.utils import historify
from timecast.utils import random


shapes = [(32, 4), (4, 32), (10,), (10, 1), (1,), (1, 10), (10, 10)]


def create_ar(shape, output_dim, history_len, history=None):
    """Create AR model"""
    model_def = AR.partial(output_dim=output_dim, history_len=history_len, history=history)
    with flax.nn.stateful() as state:
        _, params = model_def.init_by_shape(jax.random.PRNGKey(0), [shape])
        model = flax.nn.Model(model_def, params)
    return model, state


@pytest.mark.parametrize("shape", shapes)
@pytest.mark.parametrize("output_dim", [1, 10])
@pytest.mark.parametrize("history_len", [1, 10])
def test_ar_ys_shape(shape, output_dim, history_len):
    """Test output shape"""
    model, state = create_ar(shape, output_dim, history_len)
    with flax.nn.stateful(state) as _:
        ys = model(jax.random.uniform(random.generate_key(), shape=shape))

    assert ys.shape == (output_dim,)


@pytest.mark.parametrize("shape", shapes)
@pytest.mark.parametrize("output_dim", [1, 10])
@pytest.mark.parametrize("history_len", [1, 10])
def test_ar_history_shape(shape, output_dim, history_len):
    """Test history shape"""
    model, state = create_ar(shape, output_dim, history_len)
    with flax.nn.stateful(state) as state:
        _ = model(jax.random.uniform(random.generate_key(), shape=shape))

    print(state.as_dict())

    input_features = shape[0] if len(shape) == 1 else shape[1]
    assert state.as_dict()["/"]["history"].shape == (history_len, input_features)


@pytest.mark.parametrize("shape", shapes)
@pytest.mark.parametrize("output_dim", [1, 10])
@pytest.mark.parametrize("history_len", [1, 10])
def test_ar_history(shape, output_dim, history_len):
    """Test history values"""
    input_features = shape[0] if len(shape) == 1 else shape[1]
    history = jax.random.uniform(random.generate_key(), shape=(history_len, input_features))
    model, state = create_ar(shape, output_dim, history_len, history)

    onp.testing.assert_array_almost_equal(history, state.as_dict()["/"]["history"])


@pytest.mark.parametrize("shape", shapes)
@pytest.mark.parametrize("fit_intercept", [True, False])
@pytest.mark.parametrize("alpha", [0.0, 0.5, 1.0])
def test_compute_kernel_bias(shape, fit_intercept, alpha):
    """Test kernel and bias computation"""
    if len(shape) == 1:
        shape += (1,)
    X = jax.random.uniform(random.generate_key(), shape=shape)
    Y = jax.random.uniform(random.generate_key(), shape=(shape[0],))

    # Ignore underdetermined systems for now
    if shape[0] > shape[1]:
        ridge = Ridge(alpha=alpha, fit_intercept=fit_intercept)
        expected = ridge.fit(X, Y)
        kernel, bias = _compute_kernel_bias(X, Y, alpha=alpha, fit_intercept=fit_intercept)
        onp.testing.assert_array_almost_equal(expected.coef_, kernel, decimal=2)
        onp.testing.assert_array_almost_equal(expected.intercept_, bias, decimal=2)


@pytest.mark.parametrize("shape", [(40, 1), (50, 2)])
@pytest.mark.parametrize("alpha", [0.0, 0.5, 1.0])
@pytest.mark.parametrize("history_len", [1, 5, 10])
def test_ar_fit(shape, history_len, alpha):
    """Test AR class method fit"""
    X = jax.random.uniform(random.generate_key(), shape=shape)
    Y = jax.random.uniform(random.generate_key(), shape=(shape[0],))

    num_histories = shape[0] - history_len + 1
    history = historify(X, num_histories, history_len).reshape(num_histories, -1)

    kernel, bias = _compute_kernel_bias(
        history, Y[-len(history) :], alpha=alpha, fit_intercept=True
    )
    kernel = kernel.reshape(1, history_len * shape[1], 1)
    bias = jnp.expand_dims(jnp.asarray(bias), 0)

    ar, state = AR.fit(
        [(X, Y)], input_dim=X.shape[1], history_len=history_len, alpha=alpha, normalize=False
    )
    ar_kernel = ar.params["linear"]["kernel"]
    ar_bias = ar.params["linear"]["bias"]

    onp.testing.assert_array_almost_equal(kernel, ar_kernel, decimal=3)
    onp.testing.assert_array_almost_equal(bias, ar_bias, decimal=3)


def test_ar_fit_value_error():
    """Test number of observations"""
    X = jax.random.uniform(random.generate_key(), shape=(1, 10))
    Y = jax.random.uniform(random.generate_key(), shape=(10, 1))

    with pytest.raises(ValueError):
        AR.fit([(X, Y)], input_dim=10, history_len=1)


@pytest.mark.parametrize(
    "shape", [(1,), (1, 1), (10,), (10, 1), (1, 10), (10, 10), (20, 5), (5, 20)]
)
@pytest.mark.parametrize("history_len", [1, 5, 10])
@pytest.mark.parametrize("loc", [True, False])
@pytest.mark.parametrize("scale", [True, False])
def test_ar_apply(shape, history_len, loc, scale):
    """Test apply"""
    x = jax.random.uniform(random.generate_key(), shape=shape)
    if x.ndim == 1:
        x = x.reshape(1, -1)
    history = jax.random.uniform(random.generate_key(), shape=(history_len, x.shape[1]))
    kernel = jax.random.uniform(random.generate_key(), shape=(1, history_len * x.shape[1], 1))
    bias = jax.random.uniform(random.generate_key(), shape=(1,))

    loc = (
        jax.random.uniform(random.generate_key(), shape=(1, history_len * x.shape[1]))
        if loc
        else None
    )
    scale = (
        jax.random.uniform(random.generate_key(), shape=(1, history_len * x.shape[1]))
        if scale
        else None
    )

    with flax.nn.stateful() as _:
        model_def = AR.partial(
            history_len=history_len, history=history, loc=loc, scale=scale, name="AR"
        )
        _, params = model_def.init_by_shape(jax.random.PRNGKey(0), [shape])
        model = flax.nn.Model(model_def, params)
        model.params["linear"]["kernel"] = kernel
        model.params["linear"]["bias"] = bias

        result = model(x)

    expected_history = jnp.vstack((history, x))[x.shape[0] :].ravel()
    if loc is not None:
        expected_history -= loc
    if scale is not None:
        expected_history /= scale

    expected = jnp.tensordot(expected_history.reshape(1, -1), kernel) + bias

    onp.testing.assert_array_almost_equal(result, expected)



"""timecast.optim

Todo:
    * Document available optimizers
"""
import flax
from flax.optim import Adam
from flax.optim import GradientDescent
from flax.optim import LARS
from flax.optim import Momentum

from timecast.optim._adagrad import Adagrad
from timecast.optim._rmsprop import RMSProp

# from flax.optim import LAMB


@flax.struct.dataclass
class _DummyGradHyperParams:
    """DummyGrad hyper parameters"""

    add: float


class DummyGrad(flax.optim.OptimizerDef):
    """Dummy optimizer for testing"""

    def __init__(self, add: float = 0.0):
        """Initialize hyper parameters"""
        super().__init__(_DummyGradHyperParams(add))

    def init_param_state(self, param):
        """Initialize parameter state"""
        return {}

    def apply_param_gradient(self, step, hyper_params, param, state, grad):
        """Apply per-parametmer gradients"""
        new_param = param + hyper_params.add
        return new_param, state


__all__ = ["Adagrad", "Adam", "GradientDescent", "Momentum", "LAMB", "LARS", "RMSProp"]



"""flax.nn.Module for an auto-regressive online learner.

Todo:
    * Implement batching; right now, passing in a batch skips to the end and
    predicts one value, rather than predict one value per example or some sort
    of averaging
    * Implement strided history
    * Add link functions for GLM

References:
    * http://eeweb.poly.edu/iselesni/lecture_notes/least_squares/least_squares_SP.pdf
"""
from typing import Iterable
from typing import Tuple
from typing import Union

import flax
import jax
import jax.numpy as jnp
import numpy as onp

from timecast.learners.base import FitMixin
from timecast.learners.base import NewMixin
from timecast.utils import historify
from timecast.utils import internalize
from timecast.utils import random
from timecast.utils.gram import OnlineGram


def _compute_kernel_bias_gram(XTX: onp.ndarray, XTY: onp.ndarray, alpha: float = 0.0):
    """Compute linear regression parameters from gram matrix

    Notes:
        * Assumes fit_intercept=True
        * Assumes over-determined systems
    """
    reg = alpha * jnp.eye(XTX.shape[0])
    reg = jax.ops.index_update(reg, [0, 0], 0)
    beta = jnp.linalg.inv(XTX + reg) @ XTY

    return beta[1:], beta[0]


def _compute_kernel_bias(X: onp.ndarray, Y: onp.ndarray, fit_intercept=True, alpha: float = 0.0):
    """Compute linear regression parameters"""
    num_samples, num_features = X.shape

    if fit_intercept:
        if num_features >= num_samples:
            X -= X.mean(axis=0)
        X = jnp.hstack((jnp.ones((X.shape[0], 1)), X))

    reg = alpha * jnp.eye(X.shape[0 if num_features >= num_samples else 1])
    if fit_intercept:
        reg = jax.ops.index_update(reg, [0, 0], 0)

    if num_features >= num_samples:
        beta = X.T @ jnp.linalg.inv(X @ X.T + reg) @ Y
    else:
        beta = jnp.linalg.inv(X.T @ X + reg) @ X.T @ Y

    if fit_intercept:
        return beta[1:], beta[0]
    else:
        return beta, [0]


class AR(NewMixin, FitMixin, flax.nn.Module):
    """AR online learner"""

    def apply(
        self,
        x: onp.ndarray,
        history_len: int,
        output_dim: Union[Tuple[int, ...], int] = 1,
        history: onp.ndarray = None,
        loc: Union[onp.ndarray, float] = None,
        scale: Union[onp.ndarray, float] = None,
    ):
        """
        Note:
            * We expect that `x` is one- or two-dimensional
            * We reshape `x` to ensure its first axis is time and its second
              axis is input_features

        Args:
            x (onp.ndarray): input data
            history_len (int): length of AR history length
            output_dim (Union[Tuple[int, ...], int]): int or tuple
            describing output shape
            history (onp.ndarray, optional): Defaults to None. Optional
            initialization for history
            loc: mean for centering data
            scale: std for normalizing data

        Returns:
            onp.ndarray: result
        """

        if jnp.isscalar(x):
            x = [[x]]
        if x.ndim == 1:
            x = x.reshape(1, -1)

        self.history = self.state(
            "history", shape=(history_len, x.shape[1]), initializer=flax.nn.initializers.zeros
        )

        if self.is_initializing() and history is not None:
            self.history.value = jnp.vstack((self.history.value, history))[history.shape[0] :]
        elif not self.is_initializing():
            self.history.value = jnp.vstack((self.history.value, x))[x.shape[0] :]

        inputs = self.history.value.ravel().reshape(1, -1)

        if loc is not None:
            inputs -= loc

        if scale is not None:
            inputs /= scale

        y = flax.nn.DenseGeneral(
            inputs=inputs,
            features=output_dim,
            axis=(0, 1),
            batch_dims=(),
            bias=True,
            dtype=jnp.float32,
            kernel_init=flax.nn.initializers.zeros,
            bias_init=flax.nn.initializers.zeros,
            precision=None,
            name="linear",
        )
        return y

    @classmethod
    def fit(
        cls,
        data: Iterable[Tuple[onp.ndarray, onp.ndarray]],
        input_dim: int,
        history_len: int,
        output_dim: int = 1,
        normalize: bool = True,
        alpha: float = 0.0,
        key: jnp.ndarray = None,
        name: str = "ar",
        **kwargs
    ) -> flax.nn.Model:
        """Receives data as an iterable of tuples containing input time series,
        true time series

        Todo:
            * We assume input_dim is one-dimensional; we should flatten if not
            * We assume output_feature is 1, but this may not always be true
            * output_dim defaults to 1 and is ignored for now
            * Really intended for passing in timeseries at a time, not
            individual time series observations; is this the right general API?
            * Shape is (1, input_dim); what about mini-batches?

        Notes:
            * Use (1, history_len * input_dim) vectors as features (could
            consider other representations)
            * Given a time series of length N and a history of length H,
            construct N - H + 1 windows
            * We could infer input_dim from data, but for now, require
            users to explicitly provide
            * Assumes we get tuples of time series, not individual time series
            observations

        Args:
            data: an iterable of tuples containing input/truth pairs of time
            series
            input_dim: number of feature dimensions in input
            history_len: length of history to consider
            output_dim: number of feature dimensions in output
            normalize: zscore data or not
            alpha: for ridge regression
            key: random key for jax random
            name: name for the top-level module
            kwargs: Extra keyword arguments

        Returns:
            flax.nn.Model: initialized model
        """
        num_features = input_dim * history_len
        XTX = OnlineGram(num_features, normalize=normalize)
        XTY = OnlineGram(num_features, output_dim, normalize=normalize)

        for X, Y in data:
            X = internalize(X, input_dim)[0]
            Y = internalize(Y, output_dim)[0]

            if X.shape[0] != Y.shape[0]:
                raise ValueError(
                    "Input and output data must have the same" "number of observations"
                )

            # The total number of samples we will have
            num_histories = X.shape[0] - history_len + 1

            # Expand input time series X into histories, whic should result in a
            # (num_histories, history_len * input_dim)-shaped array
            history = historify(
                X, num_histories=num_histories, history_len=history_len, offset=0
            ).reshape(num_histories, -1)

            XTX.update(history)
            XTY.update(history, Y[-len(history) :])

        if XTX.observations == 0:
            raise IndexError("No data to fit")

        kernel, bias = _compute_kernel_bias_gram(XTX.matrix, XTY.matrix, alpha=alpha)
        kernel = kernel.reshape(1, num_features, 1)

        loc = XTX.mean if normalize else None
        scale = XTX.std if normalize else None

        model_def = AR.partial(
            history_len=history_len, output_dim=output_dim, loc=loc, scale=scale, name=name
        )

        if key is None:
            key = random.generate_key()

        with flax.nn.stateful() as state:
            _, params = model_def.init_by_shape(key, [(1, input_dim)])

        model = flax.nn.Model(model_def, params)
        model.params["linear"]["kernel"] = kernel
        model.params["linear"]["bias"] = bias

        return model, state



"""Objective functions for composing loss functions for multiple learners"""
import jax


def residual(x, y, loss_fn, model):
    """List of learners where each learner trains on the residual of the previous"""
    y_hats = model(x)
    target, y_hat, loss = y, y_hats[0], 0.0
    for i in range(len(y_hats) - 1):
        loss += loss_fn(target - y_hats[i], y_hats[i + 1])
        target -= y_hats[i]
        y_hat += y_hats[i + 1]
    return loss, y_hat


def xboost(x, y, loss_fn, model):
    """List of learners governed by xboost

    Notes:
        * See: https://arxiv.org/pdf/1906.08720.pdf
    """
    y_hats = model(x)
    g = jax.grad(loss_fn)
    u, loss = 0.0, 0.0
    for i in range(len(y_hats)):
        eta = 2.0 / (i + 2.0)
        loss += g(y, u) * y_hats[i] + (1.0 / 2.0) * y_hats[i] * y_hats[i]
        u = (1.0 - eta) * u + eta * y_hats[i]
    return loss.reshape(()), u



"""Test combinators"""
import flax
import jax
import numpy as onp
import pytest

from timecast.learners import Parallel
from timecast.learners import Sequential
from timecast.utils import random

shapes = [(4, 32), (10,), (10, 1), (1,), (1, 10)]


class Identity(flax.nn.Module):
    """Identity layer"""

    def apply(self, x: onp.ndarray):
        """
        Args:
            x (onp.ndarray): input data
        Returns:
            onp.ndarray: result
        """
        return x


class Plus(flax.nn.Module):
    """Adds constant"""

    def apply(self, x, z):
        """
        Args:
            x (onp.ndarray): input data
        Returns:
            onp.ndarray: result
        """
        return x + z


@pytest.mark.parametrize("shape", shapes)
def test_parallel(shape):
    """Test Parallel combinator"""
    model_def = Parallel.partial(learners=[Identity, Identity])
    _, params = model_def.init_by_shape(jax.random.PRNGKey(0), [shape])
    model = flax.nn.Model(model_def, params)

    X = jax.random.uniform(random.generate_key(), shape=(shape))
    ys = model(X)

    for y in ys:
        onp.testing.assert_array_almost_equal(X, y)


@pytest.mark.parametrize("shape", shapes)
def test_sequential(shape):
    """Test Sequential combinator"""
    model_def = Sequential.partial(learners=[Identity, Plus.partial(z=2)])
    _, params = model_def.init_by_shape(jax.random.PRNGKey(0), [shape])
    model = flax.nn.Model(model_def, params)

    X = jax.random.uniform(random.generate_key(), shape=shape)
    ys = model(X)

    onp.testing.assert_array_almost_equal(X + 2, ys)



"""flax.nn.Module for predicting a constant value

Todo:
    * Validate data and add negative tests
"""
from numbers import Real
from typing import Union

import flax
import jax
import jax.numpy as np
import numpy as onp

from timecast.learners.base import NewMixin


class PredictConstant(NewMixin, flax.nn.Module):
    """Constant online learner"""

    def apply(self, x: onp.ndarray, c: Union[onp.ndarray, Real] = 0):
        """
        Note:
            * Returns `c` in the shape of `x` as the prediction for the next time step

        Args:
            x (onp.ndarray): input data
            c (Real): prediction

        Returns:
            onp.ndarray: result

        Raises:
            ValueError: if `c` is not a scalar or does not match the shape of `x`
        """
        # TODO (flax): Remove this once flax updates
        _ = self.param("dummy", (), jax.nn.initializers.zeros)

        if np.isscalar(c):
            return np.ones_like(x) * c

        if c.shape != x.shape:
            raise ValueError("Constant must be scalar or match input shape")

        return c



"""timecast.series.uci"""
import os
from typing import Tuple

import numpy as onp

from timecast.series._core import generate_timeline


def generate(path=None) -> Tuple[onp.ndarray, onp.ndarray]:
    """
    Description: Outputs various weather metrics from a UCI dataset from 13/3/2012 to 11/4/2012

    References:
        * https://archive.ics.uci.edu/ml/machine-learning-databases/00274/NEW-DATA.zip
    """

    return generate_timeline(
        path=path or os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../data/uci.txt"),
        name="7:CO2_Habitacion_Sensor",
        delimiter=" ",
    )



"""timecast.optim._adagrad"""
import flax
import jax.numpy as jnp
import numpy as onp


@flax.struct.dataclass
class _AdagradHyperParams:
    """Adagrad hyper parameters"""

    learning_rate: float
    eps: float


@flax.struct.dataclass
class _AdagradParamState:
    """Adagrad parameter state"""

    G: onp.ndarray


class Adagrad(flax.optim.OptimizerDef):
    """Adagrad optimizer"""

    def __init__(self, learning_rate: float = 1.0, eps=1e-8):
        """Initialize hyper parameters"""
        hyper_params = _AdagradHyperParams(learning_rate, eps)
        super().__init__(hyper_params)

    def init_param_state(self, param):
        """Initialize parameter state"""
        return _AdagradParamState(jnp.zeros_like(param))

    def apply_param_gradient(self, step, hyper_params, param, state, grad):
        """Apply per-parameter gradients"""

        new_G = state.G + jnp.square(grad)
        new_param = param - hyper_params.learning_rate * grad / jnp.sqrt(new_G + hyper_params.eps)
        new_state = _AdagradParamState(new_G)

        return new_param, new_state



"""timecacst.api: testing"""
import jax
import numpy as onp

from timecast.api import smap
from timecast.learners import AR
from timecast.learners import PredictLast
from timecast.optim import DummyGrad
from timecast.utils import random


def test_smap_no_objective():
    """Test smap without objective"""
    optimizer_def = DummyGrad(add=4.0)
    model, state = PredictLast.new((1, 10))
    optimizer = optimizer_def.create(model)

    X = jax.random.uniform(random.generate_key(), shape=(5, 10))
    Y = jax.random.uniform(random.generate_key(), shape=(5, 10))
    pred, optimizer, state = smap(X, Y, optimizer)

    onp.testing.assert_array_equal(X, pred)

    # TODO (flax): this will fail once we remove dummy
    assert 20 == optimizer.target.params["dummy"]


def test_smap_state():
    """Test smap with stateful learner"""
    optimizer_def = DummyGrad(add=4.0)
    model, state = AR.new((1, 10), output_dim=1, history_len=1)
    optimizer = optimizer_def.create(model)

    X = jax.random.uniform(random.generate_key(), shape=(5, 10))
    Y = jax.random.uniform(random.generate_key(), shape=(5, 10))
    pred, optimizer, state = smap(X, Y, optimizer, state=state)

    onp.testing.assert_array_equal(state.state["/"]["history"], X[-1, :].reshape(1, -1))
    onp.testing.assert_array_equal(optimizer.target.params["linear"]["bias"], 20.0)



"""timecast top-level API"""
from functools import partial
from typing import Callable
from typing import Tuple
from typing import Union

import flax
import jax
import jax.numpy as jnp
import numpy as onp


def _objective(x, y, loss_fn, model):
    """Default objective function"""
    y_hat = model(x)
    return loss_fn(y, y_hat), y_hat


def smap(
    X: Union[onp.ndarray, Tuple[onp.ndarray, ...]],
    Y: Union[onp.ndarray, Tuple[onp.ndarray, ...]],
    optimizer: flax.optim.Optimizer,
    loss_fn: Callable[[onp.ndarray, onp.ndarray], onp.ndarray] = lambda true, pred: jnp.square(
        true - pred
    ).mean(),
    state: flax.nn.base.Collection = None,
    objective: Callable[
        [
            onp.ndarray,
            onp.ndarray,
            Callable[[onp.ndarray, onp.ndarray], onp.ndarray],
            flax.nn.base.Model,
        ],
        Tuple[onp.ndarray, onp.ndarray],
    ] = None,
):
    """Take gradients steps performantly on one data item at a time

    Args:
        X: onp.ndarray or tuple of onp.ndarray of inputs
        Y: onp.ndarray or tuple of onp.ndarray of outputs
        optimizer: initialized optimizer
        loss_fn: loss function to compose where first arg is true value and
        second is pred
        state: state required by flax
        objective: function composing loss functions

    Returns:
        onp.ndarray: result
    """
    state = state or flax.nn.Collection()
    objective = objective or _objective

    def _smap(optstate, xy):
        """Helper function"""
        x, y = xy
        optimizer, state = optstate
        func = partial(objective, x, y, loss_fn)
        with flax.nn.stateful(state) as state:
            (loss, y_hat), grad = jax.value_and_grad(func, has_aux=True)(optimizer.target)
        return (optimizer.apply_gradient(grad), state), y_hat

    (optimizer, state), pred = jax.lax.scan(_smap, (optimizer, state), (X, Y))
    return pred, optimizer, state



"""flax.nn.Module for a principal component regression online learner."""
from typing import Iterable
from typing import Tuple
from typing import Union

import flax
import jax.numpy as jnp
import numpy as onp

from timecast.learners._ar import _compute_kernel_bias_gram
from timecast.learners.base import FitMixin
from timecast.learners.base import NewMixin
from timecast.utils import historify
from timecast.utils import internalize
from timecast.utils import random
from timecast.utils.gram import OnlineGram


def _compute_pca_projection(X: onp.ndarray, k: int, center=False) -> onp.ndarray:
    """Compute PCA projection"""

    if center:
        X = X - X.mean(axis=0)
        # X /= X.shape[0] - 1

    # Compute SVD
    # X: (steps - history_len + 1, history_len * input_dim) -> (H, d)
    # U: (H, H), but because full_matrices=False, (H, d)
    # S: (min(H, d),)
    # VT: (d, d)
    U, S, VT = jnp.linalg.svd(X, full_matrices=False, compute_uv=True)

    # Get index of top K eigen values
    top_k = (-jnp.square(S)).argsort()[:k]

    # Get projection
    # projection: (d, k)
    projection = VT[top_k].T

    return projection


class PCR(NewMixin, FitMixin, flax.nn.Module):
    """PCR online learner"""

    def apply(
        self,
        x: onp.ndarray,
        history_len: int,
        projection: onp.ndarray,
        output_dim: Union[Tuple[int, ...], int] = 1,
        history: onp.ndarray = None,
        loc: Union[onp.ndarray, float] = None,
        scale: Union[onp.ndarray, float] = None,
    ):
        """
        Todo:
            * AR doesn't take any history
        output_dim (Union[Tuple[int, ...], int]): int or tuple describing
        output shape

        Note:
            * We expect that `x` is one- or two-dimensional
            * We reshape `x` to ensure its first axis is time and its second
              axis is input_features

        Args:
            x (onp.ndarray): input data
            history_len (int): length of AR history length
            output_dim (Union[Tuple[int, ...], int]): int or tuple
            describing output shape
            history (onp.ndarray, optional): Defaults to None. Optional
            initialization for history
            loc: mean for centering data
            scale: std for normalizing data

        Returns:
            onp.ndarray: result
        """
        if jnp.isscalar(x):
            x = [[x]]
        if x.ndim == 1:
            x = x.reshape(1, -1)

        self.history = self.state(
            "history", shape=(history_len, x.shape[1]), initializer=flax.nn.initializers.zeros
        )

        if self.is_initializing() and history is not None:
            self.history.value = jnp.vstack((self.history.value, history))[history.shape[0] :]
        elif not self.is_initializing():
            self.history.value = jnp.vstack((self.history.value, x))[x.shape[0] :]

        inputs = self.history.value.reshape(1, -1) @ projection

        if loc is not None:
            inputs -= loc

        if scale is not None:
            inputs /= scale

        y = flax.nn.DenseGeneral(
            inputs=inputs,
            features=output_dim,
            axis=(0, 1),
            batch_dims=(),
            bias=True,
            dtype=jnp.float32,
            kernel_init=flax.nn.initializers.zeros,
            bias_init=flax.nn.initializers.zeros,
            precision=None,
            name="linear",
        )
        return y

    @classmethod
    def fit(
        cls,
        data: Iterable[Tuple[onp.ndarray, onp.ndarray]],
        input_dim: int,
        history_len: int,
        output_dim: int = 1,
        k: int = None,
        normalize: bool = True,
        alpha: float = 0.0,
        key: jnp.ndarray = None,
        name: str = "pcr",
        **kwargs
    ) -> flax.nn.Model:
        """Receives data as an iterable of tuples containing input time series,
        true time series

        Todo:
            * We assume input_dim is one-dimensional; we should flatten if not
            * We assume output_feature is 1, but this may not always be true
            * output_dim defaults to 1 and is ignored for now
            * Really intended for passing in timeseries at a time, not
            individual time series observations; is this the right general API?
            * Shape is (1, input_dim); what about mini-batches?

        Notes:
            * Use (1, history_len * input_dim) vectors as features (could
            consider other representations)
            * Ignore true value (i.e., look at features only) (should we
            consider impact of features on dependent variable?)
            * Given a time series of length N and a history of length H,
            construct N - H + 1 windows
            * We could infer input_dim from data, but for now, require
            users to explicitly provide
            * Assumes we get tuples of time series, not individual time series
            observations

        Args:
            data: an iterable of tuples containing input/truth pairs of time
            series
            input_dim: number of feature dimensions in input
            history_len: length of history to consider
            output_dim: number of feature dimensions in output
            k: number of PCA components to keep. Default is min(num_histories,
            normalize: zscore data or not
            input_dim)
            alpha: Parameter to pass to ridge regression for AR fit
            key: random key for jax random
            name: name for the top-level module
            kwargs: Extra keyword arguments

        Returns:
            flax.nn.Model: initialized model
        """
        num_features = input_dim * history_len
        XTX = OnlineGram(num_features, normalize=normalize)
        XTY = OnlineGram(num_features, output_dim, normalize=normalize)

        for X, Y in data:
            X = internalize(X, input_dim)[0]
            Y = internalize(Y, output_dim)[0]

            if X.shape[0] != Y.shape[0]:
                raise ValueError(
                    "Input and output data must have the same" "number of observations"
                )

            # The total number of samples we will have
            num_histories = X.shape[0] - history_len + 1

            # Expand input time series X into histories, whic should result in a
            # (num_histories, history_len * input_dim)-shaped array
            history = historify(
                X, num_histories=num_histories, history_len=history_len, offset=0
            ).reshape(num_histories, -1)

            XTX.update(history)
            XTY.update(history, Y[-len(history) :])

        if XTX.observations == 0:
            raise IndexError("No data to fit")

        # Compute k
        k = num_features if k is None else min(k, num_features)
        # TODO: check
        projection = _compute_pca_projection(XTX.matrix, k)

        # TODO: check
        # X: (n, d)
        # XTX.matrix: (d, d)
        # projection: (d, k)
        # X @ projection: (n, k)
        # (X @ projection).T @ (X @ projection): (k, k)
        # projection.T @ X.T @ X @ projection = (k, d) @ (d, n) @ (n, d) @ (d, k)
        kernel, bias = _compute_kernel_bias_gram(
            projection.T @ XTX.matrix @ projection,
            projection.T @ XTY.matrix,
            fit_intercept=True,
            alpha=alpha,
        )
        kernel = kernel.reshape(1, k, 1)

        loc = XTX.mean if normalize else None
        scale = XTX.std if normalize else None

        model_def = PCR.partial(
            history_len=history_len,
            projection=projection,
            output_dim=output_dim,
            loc=loc,
            scale=scale,
            name=name,
        )

        if key is None:
            key = random.generate_key()

        with flax.nn.stateful() as state:
            _, params = model_def.init_by_shape(key, [(1, input_dim)])

        model = flax.nn.Model(model_def, params)
        model.params["linear"]["kernel"] = kernel
        model.params["linear"]["bias"] = bias

        return model, state



"""timecast.learners.Index: testing"""
import flax
import jax
import numpy as onp
import pytest

from timecast.learners import Index
from timecast.utils import random

shapes = [(4, 32), (10,), (10, 1), (1,), (1, 10)]


def create_index(shape, index):
    """Generate Index model and state"""
    with flax.nn.stateful() as state:
        model_def = Index.partial(index=index)
        _, params = model_def.init_by_shape(jax.random.PRNGKey(0), [shape])
        model = flax.nn.Model(model_def, params)
    return model, state


@pytest.mark.parametrize("shape", shapes)
def test_index(shape):
    """Test Index"""
    model, state = create_index(shape, index=0)

    X = jax.random.uniform(random.generate_key(), shape=shape)
    with flax.nn.stateful(state) as state:
        ys = model(X)

    onp.testing.assert_array_almost_equal(X[0], ys)


@pytest.mark.parametrize("shape", shapes)
def test_index_bad_index(shape):
    """Test bad_index"""
    with pytest.raises(IndexError):
        _, _ = create_index(shape, index=-1)


def test_index_scalar():
    """Test scalar input"""
    model, state = create_index((2, 1), index=0)
    with pytest.raises(ValueError):
        with flax.nn.stateful(state) as state:
            _ = model(1)



"""timecast.optim._rmsprop"""
import flax
import jax.numpy as jnp
import numpy as onp


@flax.struct.dataclass
class _RMSPropHyperParams:
    """RMSProp hyper parameters"""

    learning_rate: float
    beta2: float
    eps: float


@flax.struct.dataclass
class _RMSPropParamState:
    """RMSProp parameter state"""

    v: onp.ndarray


class RMSProp(flax.optim.OptimizerDef):
    """RMSProp optimizer"""

    def __init__(self, learning_rate: float = 1.0, beta2=0.999, eps=1e-8):
        """Initialize hyper parameters"""
        hyper_params = _RMSPropHyperParams(learning_rate, beta2, eps)
        super().__init__(hyper_params)

    def init_param_state(self, param):
        """Initialize parameter state"""
        return _RMSPropParamState(jnp.zeros_like(param))

    def apply_param_gradient(self, step, hyper_params, param, state, grad):
        """Apply per-parameter gradients"""

        new_v = hyper_params.beta2 * state.v + (1.0 - hyper_params.beta2) * jnp.square(grad)
        new_param = param - hyper_params.learning_rate * grad / (jnp.sqrt(new_v) + hyper_params.eps)
        new_state = _RMSPropParamState(new_v)

        return new_param, new_state



"""timecast.series.arma"""
from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as onp


def generate(
    n: int = 1000, p: int = 3, q: int = 3, dim: int = 1, c: float = 0.0, mag: float = 0.1,
) -> Tuple[onp.ndarray, onp.ndarray]:
    """
    Description: Randomly initialize the hidden dynamics of the system.
    Args:
        n (int): number of values to generate
        p (int): Autoregressive dynamics. If type int
        then randomly initializes a Gaussian length-p vector with L1-norm
        bounded by 1.0.  If p is a 1-dimensional numpy.ndarray then uses it
        as dynamics vector.
        q (int): Moving-average dynamics. If type int then
        randomly initializes a Gaussian length-q vector (no bound on norm).
        If p is a 1-dimensional numpy.ndarray then uses it as dynamics
        vector.
        dim (int): Dimension of values.
        c (float): Default value follows a normal distribution. The ARMA
        dynamics follows the equation x_t = c + AR-part + MA-part + noise,
        and thus tends to be centered around mean c.
        mag (float): Noise magnitude
    Returns:
        Tuple[onp.ndarray, onp.ndarray]: X, y pair
    """
    phi = jnp.asarray(onp.random.rand(p))
    phi = 0.99 * phi / jnp.linalg.norm(phi, ord=1)

    psi = jnp.asarray(onp.random.rand(q))

    c = jnp.asarray(onp.random.rand(dim)) if c is None else c
    x = jnp.asarray(onp.random.rand(p, dim))

    noise = mag * jnp.asarray(onp.random.rand(q, dim))

    def step(carry, eps):
        """Internal step function for ARMA"""
        x, noise = carry
        x_ar = jnp.dot(x.T, phi)

        x_ma = jnp.dot(noise.T, psi)

        y = c + x_ar + x_ma + eps
        next_x = jnp.roll(x, n)
        next_noise = jnp.roll(noise, n)

        next_x = jax.ops.index_update(next_x, 0, y)
        next_noise = jax.ops.index_update(next_noise, 0, eps)

        return (next_x, next_noise), y

    EPS = jnp.asarray(onp.random.rand(n + 1, dim))

    _, y = jax.lax.scan(step, (x, noise), EPS)

    return y[:-1].reshape(-1, 1), y[1:].reshape(-1, 1)



"""timecast.series.sp500"""
import os
from typing import Tuple

import numpy as onp

from timecast.series._core import generate_timeline


def generate(path=None) -> Tuple[onp.ndarray, onp.ndarray]:
    """
    Description: Outputs the log-normalized daily opening price of the S&P 500
    stock market index from January 3, 1986 to June 29, 2018.
    """

    return generate_timeline(
        path or os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../data/sp500.csv")
    )



"""timecast.series.rnn"""
from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as onp


def generate(
    n: int = 1000, input_dim: int = 5, output_dim: int = 1, hidden_dim: int = 64
) -> Tuple[onp.ndarray, onp.ndarray]:
    """
    Description: output from a randomly initialized recurrent neural network
    """
    W_h = jnp.asarray(onp.random.rand(hidden_dim, hidden_dim))
    W_x = jnp.asarray(onp.random.rand(hidden_dim, input_dim))
    W_out = jnp.asarray(onp.random.rand(output_dim, hidden_dim))
    b_h = jnp.zeros(hidden_dim)
    hidden = jnp.zeros(hidden_dim)

    def step(h, x):
        """Internal function for RNN step"""
        hid = jnp.tanh(jnp.dot(W_h, h) + jnp.dot(W_x, x) + b_h)
        y = jnp.dot(W_out, hid)
        return hid, y

    X = jnp.asarray(onp.random.rand(n, input_dim))
    hidden, y = jax.lax.scan(step, hidden, X)

    return X, y.reshape(-1, 1)



"""timecast.utils.historify"""
import jax
import numpy as onp
import pytest

from timecast.utils import historify
from timecast.utils import random


@pytest.mark.parametrize("m", [1, 10])
@pytest.mark.parametrize("n", [1, 10])
@pytest.mark.parametrize("num_histories", [0, 1, 10])
@pytest.mark.parametrize("history_len", [-1, 0, 1, 10])
@pytest.mark.parametrize("offset", [0, 1, 10])
def test_historify(m, n, num_histories, history_len, offset):
    """Test history-making"""
    X = jax.random.uniform(random.generate_key(), shape=(m, n))

    if (
        num_histories < 1
        or history_len < 1
        or X.shape[0] < offset + num_histories + history_len - 1
    ):
        with pytest.raises(ValueError):
            historify(X, num_histories, history_len, offset)

    else:
        batched = historify(X, num_histories, history_len, offset)

        for i, batch in enumerate(batched):
            onp.testing.assert_array_almost_equal(batch, X[i + offset : i + offset + history_len])



"""timecast.series.crypto"""
import os
from typing import Tuple

import jax.numpy as jnp
import numpy as onp
import pandas as pd


def generate(path=None) -> Tuple[onp.ndarray, onp.ndarray]:
    """
    Description: Outputs the daily price of bitcoin from 2013-04-28 to 2018-02-10
    """

    data = pd.read_csv(
        path or os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../data/crypto.csv")
    )

    data = jnp.asarray(data[data.Currency == "bitcoin"].Price)

    return data[:-1].reshape(-1, 1), data[1:].reshape(-1, 1)



"""Experiment decorator for running experiments in parallel"""
import inspect
from collections.abc import Iterable
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import jax.numpy as jnp
import multiprocess.context as ctx
from pathos.pools import ProcessPool

ctx._force_start_method("spawn")


def runner(args: Tuple[Callable, Dict[Any, Any]]) -> Any:
    """Runner function for process pool
    
    Args:
        args: A pair where the first element is the function to run and
        the second element is a dictionary of kwargs

    Returns:
        Any: the result of the function in the first element
    """
    return args[0](**args[1])


def product(*args: Any) -> List[Any]:
    """Recursive function to yield the cartesian product of a list of arguments

    Notes:
        * We don't use itertools.product because it will recursively examine its
        arguments and we want a tuple to stay a tuple in the resulting product
    """
    if args:
        for a in args[0]:
            for prod in product(*args[1:]) if args[1:] else ((),):
                yield (a,) + prod


class experiment:
    """Class decorator to run experiments"""

    def __init__(self, argnames: Union[List[str], str], arglists: List[Any]) -> None:
        """Initializes the experiment with argnames and arglists

        Args:
            argnames: List of argument names or comma-delimited string of names (as in pytest)
            arglists: A list of one or more tuples of values for each name in argnames
        """
        if isinstance(argnames, list):
            self._argnames = argnames
        else:
            self._argnames = [argname.strip() for argname in argnames.split(",")]

        # Validate that all arglists have the correct number of arguments
        for arglist in arglists:
            arglist_length = 1 if "__len__" not in dir(arglist) else len(arglist)
            if arglist_length != len(self._argnames):
                raise ValueError(
                    (
                        "Each arglist must have the same length as provided argnames. "
                        "Got names {} and list {}"
                    ).format(argnames, arglist)
                )

        self._arglists = arglists
        self._spec = [(self._argnames, self._arglists)]

    def __call__(self, funcOrExp):
        """Decorator magic

        Notes:
            * Takes either a Callable or experiment object so we can chain decorators
        """

        # If we see an experiment, pass on the function to execute
        if hasattr(funcOrExp, "_func"):
            self._func = funcOrExp._func
            self._spec.extend(funcOrExp._spec)

        # If we see a callable, we know we're wrapping a naked function
        elif isinstance(funcOrExp, Callable):
            self._func = funcOrExp

        return self

    def _validate(self):
        """Validate an experiment's argnames and arglists against _func's arguments
        
        Notes:
            * Each decorated call to @experiment must specify arguments disjoint with
            every other call to @experiment (i.e., can't have a call "a,b,c" and "b")
            * The union of all arguments specified must be a subset of _func's arguments
            and a superset of _func's arguments without defaults
            * If _func has *args, we ignore
            * If _func has **kwargs, it will eat up any unused arguments
        """
        # Check for duplicate argument specifications
        argnames = set()
        for spec in self._spec:
            for argname in spec[0]:
                if argname in argnames:
                    raise ValueError("Found duplicate argname {} in {}".format(argname, spec))
                argnames.add(argname)

        func_fullargspec = inspect.getfullargspec(self._func)
        func_args = func_fullargspec.args
        func_args_without_defaults = (
            func_args
            if func_fullargspec.defaults is None
            else func_args[: -len(func_fullargspec.defaults)]
        )

        # NOTE: We don't use set operations because we want more descriptive error messages

        # Find any missing arguments
        missing_args = []
        for arg in func_args_without_defaults:
            if arg not in argnames:
                missing_args.append(arg)

        if len(missing_args) > 0:
            raise ValueError(
                "Arguments without defaults not found. Required: {}. Supplied: {}".format(
                    missing_args, argnames
                )
            )

        # Find any extra arguments
        extra_args = []
        if func_fullargspec.varkw is None:
            for arg in argnames:
                if arg not in func_args:
                    extra_args.append(arg)

        if len(extra_args) > 0:
            raise ValueError("Found unused arguments: {}".format(extra_args))

    def _generate_arglists(self):
        """Create a generator to pass function and arguments to a multiprocessing pool"""
        argnames = [item for sublist in [spec[0] for spec in self._spec] for item in sublist]
        arglists = product(*[spec[1] for spec in self._spec])
        for arglist in arglists:
            flattened = []

            # Flatten a list of lists and atoms
            for arg in arglist:
                if isinstance(arg, Iterable):
                    flattened.extend(arg)
                else:
                    flattened.append(arg)
            yield (self._func, {key: val for key, val in zip(argnames, flattened)})

    def run(self, processes=1, chunksize=1, tqdm=None):
        """Execute the experiment"""
        self._validate()

        num_jobs = jnp.product([len(spec[1]) for spec in self._spec])

        single = lambda: map(runner, self._generate_arglists())  # noqa: E731

        # TODO: Figure out why parallel hangs in pytest
        parallel = lambda: ProcessPool(nodes=processes).imap(  # noqa: E731
            runner, self._generate_arglists(), chunksize=chunksize
        )
        process = single if processes == 1 else parallel
        return list(process() if tqdm is None else tqdm.tqdm(process(), total=num_jobs))



"""timecast.utils.externalize: testing"""
import jax.numpy as jnp
import numpy as onp
import pytest

from timecast.utils import externalize


@pytest.mark.parametrize("shape", [(), (1,), (1, 2, 3)])
def test_externalize_type_error(shape):
    """Test type error"""
    with pytest.raises(TypeError):
        externalize(jnp.ones(shape), dim=1)


def test_externalize_not_jnp_ndarray():
    """Test instance error"""
    with pytest.raises(TypeError):
        externalize(1, dim=1)


@pytest.mark.parametrize("dim", [-2, 3])
def test_externalize_dim_value_error(dim):
    """Test value error"""
    with pytest.raises(ValueError):
        externalize(jnp.ones((1, 1)), dim)


@pytest.mark.parametrize("shape", [(1, 10), (10, 1), (2, 10), (10, 2), (1, 1), (10, 10)])
@pytest.mark.parametrize("dim", [-1, 0, 1, 2])
@pytest.mark.parametrize("was_jax", [True, False])
def test_externalize(shape, dim, was_jax):
    """Test externalize"""
    X = jnp.ones(shape)

    if dim == 2:
        onp.testing.assert_array_equal(X, externalize(X, dim, was_jax))
    elif dim == -1 and len(X.ravel()) == 1:
        assert 1 == externalize(X, dim)
    elif dim == 1 and len(X.ravel()) == 1:
        assert jnp.ones(1) == externalize(X, dim)
    else:
        onp.testing.assert_array_equal(X.squeeze(), externalize(X, dim, was_jax))



"""flax.nn.Module for taking an index from input"""
import flax
import jax.numpy as jnp
import numpy as onp


class Index(flax.nn.Module):
    """Identity index online learner"""

    def apply(self, x: onp.ndarray, index: int):
        """
        Note:
            * Returns `x[index]` as the prediction for the next time step
            * This is a workaround for the case where we have a blackbox series
            of predictions (see documentation)

        Args:
            x (onp.ndarray): input data
            index (int): index to take

        Returns:
            onp.ndarray: result
        """

        if jnp.isscalar(x):
            raise ValueError("Input x must be an array for Index learner")

        # We don't check for index < x.shape[0] because this confuses flax's
        # init_by_shape
        if not isinstance(index, int) or index < 0:
            raise IndexError("Index must be positive. Got index {}".format(index))

        return x[index]



"""flax.nn.Module for predicting last value

Todo:
    * Implement last value n steps ago (requires state)
"""
import flax
import jax
import numpy as onp

from timecast.learners.base import NewMixin


class PredictLast(NewMixin, flax.nn.Module):
    """Identity online learner"""

    def apply(self, x: onp.ndarray):
        """
        Note:
            * Returns `x` as the prediction for the next time step

        Args:
            x (onp.ndarray): input data

        Returns:
            onp.ndarray: result
        """
        # TODO (flax): Remove this once flax updates
        _ = self.param("dummy", (), jax.nn.initializers.zeros)
        return x



"""Utility functions"""
from numbers import Real
from random import randint
from typing import Tuple
from typing import Union

import jax
import jax.numpy as jnp
import numpy as onp


def historify(X: onp.ndarray, num_histories: int, history_len: int, offset: int = 0):
    """Generate (num_histories, history_len, feature_dim) history from time series data

    Todo:
        * Implement striding

    Args:
        X: 2D array of features organized as (time, feature_dim)
        num_histories: number of history windows to create
        history_len: length of a history window
        offset: how many time steps to offset

    Returns:
        onp.ndarray: 3D array organized as (num_histories, history_len,
        feature_dim)
    """
    if num_histories < 1 or history_len < 1:
        raise ValueError("Must have positive history_len and at least one window")
    if X.shape[0] < offset + num_histories + history_len - 1:
        raise ValueError(
            "Not enough history ({}) to produce {} windows of length {} with offset {}".format(
                X.shape[0], num_histories, history_len, offset
            )
        )
    return jnp.swapaxes(
        jnp.stack([jnp.roll(X, shift=-(i + offset), axis=0) for i in range(history_len)]), 0, 1
    )[:num_histories]


def internalize(
    x: Union[onp.ndarray, jnp.ndarray, Real], dim: int
) -> Tuple[bool, jnp.ndarray, int]:
    """ Takes user input data and converts into a consistent internal form
    Args:
        x (Union[onp.ndarray, jnp.ndarray, Real]): some input data to be
        converted dim (int): The input dimension specified at
        instantiation
    Notes:
        * Inputs can be scalar, 0D, 1D, or 2D.
        * While internally, we may have >2D arrays, we consider that an error
        if coming from the user
        * Internal representation is 2D array where dimension 0 is time and
        dimension 1 is feature
        * Inputs can be single time series values (value) or a sequence of time
        series values (batch)
        * Scalar and 0D inputs are values if dim is 1
        * 1D input is a value if its length matches the dim
        * 1D input is a batch if dim is 1 (previous rule takes
        precedence)
        * 2D input is a value if dimension 0 is 1 and dimension 1 is dim
        * 2D input is a batch if dimension 1 equals dim
    Warning:
        * When x is 2D, dimensions are always (time, feature)
        * We don't check whether the NumPy array is well-formed or not (e.g.,
        single data type, np.product(x.shape) == np.len(x.ravel()))
    Returns:
        x (jnp.ndarray): internalized version of input
        is_value (bool): whether the input was a value or a batch
        dim (int): number of dimensions of original input. -1 means
        np.isscalar(x) is True
        was_jax (bool): whether or not the input array was a
        jax.numpy.DeviceArray
    Raises:
        TypeError: x is not NumPy array or real number, or has more than two
        dimensions
        ValueError: input x does not match dim
    """
    was_jax = isinstance(x, jnp.DeviceArray)
    is_scalar, x = jnp.isscalar(x), jnp.asarray(x)

    if x.ndim > 2:
        raise TypeError("Input cannot have more than two dimensions")

    # -1 if scalar, otherwise take original NumPy dimension
    ndim = -1 if is_scalar else x.ndim

    # If Python or 0D NumPy array, x is a value
    if ndim <= 0 and dim == 1:
        return jnp.asarray([[x]]), True, ndim, was_jax

    # If 1D array and length is equal to dim, x is a value
    if ndim == 1 and len(x.ravel()) == dim:
        return x.reshape(1, -1), True, ndim, was_jax

    # If 1D array and dim is 1, x is a batch
    if ndim == 1 and dim == 1:
        return x.reshape(-1, 1), False, ndim, was_jax

    # If 2D array and ndimension 1 is equal to dim, x is a batch
    if ndim == 2 and x.shape[1] == dim:
        return x, x.shape[0] == 1, ndim, was_jax

    raise ValueError("Input shape {} does not match dim {}".format(x.shape, dim))


def externalize(x: jnp.ndarray, dim: int, was_jax: bool = False) -> Union[onp.ndarray, jnp.ndarray]:
    """Takes internal representation of 2D (time, feature) and returns to user
    Args:
        x (Union[jnp.ndarray, Real]): some input data to be converted
        dim (int): The original number of dimensions (-1 for scalar, n for nD
        NumPy array)
        was_jax (bool): whether original array was jax.numpy.DeviceArray
    Warning:
        * Transformations may not always be from R^n to R^n, so we have to make
        some guesses
        * We don't check whether the NumPy array is well-formed or not (e.g.,
        single data type, jnp.product(x.shape) == jnp.len(x.ravel()))
    Notes:
        * We assume if the original input was less than 2, we should squeeze
        out as many dimensions as we can
        * We leave input as is if original dimension was 2
        * It's not clear when to use this, especially if users are not
        consistent with their inputs
    Returns:
        x (Union[onp.ndarray, jnp.ndarray]): externalized version of input
    Raises:
        TypeError: x is not a NumPy array with two dimensions
        ValueError: dim is not -1, 0, 1, or 2; input x does not match dim
        or dim
    """
    if not isinstance(x, jnp.ndarray):
        raise TypeError("x has incorrect type {}".format(type(x)))

    if was_jax:
        x = onp.asarray(x)
    else:
        x = jnp.asarray(x)

    if x.ndim != 2:
        raise TypeError("x does not have two dimensions")
    if dim < -1 or dim > 2:
        raise ValueError("Original dimension must be -1, 0, 1, or 2")

    # If the original dimension was 2, just return
    if dim == 2:
        return x

    # Need to be careful with x with one item
    if len(x.ravel()) == 1:

        # If the original was a scalar, return a scalar
        if dim == -1:
            return x.item()

        # If the original was 1D, return 1D
        if dim == 1:
            return x.ravel()

    # Otherwise, try to squeeze as many dimensions as we can
    if dim < 2:
        return x.squeeze()


class random:
    """Singleton class to manage global jax randomness keys"""

    class __random:
        """Nested class to define instance"""

        def __init__(self) -> None:
            """Initialize instance"""
            self._key = jax.random.PRNGKey(0)

        def set_key(self, seed=None) -> None:
            """Fix global random key to ensure reproducibility of results.
            Args:
                seed (int): key that determines reproducible output
            """
            if seed is None:
                seed = randint(1, 9223372036854775807 + 1)
            assert type(seed) == int
            self._key = jax.random.PRNGKey(seed)

        def get_key(self) -> jnp.ndarray:
            """Return current global key"""
            return self._key

        def generate_key(self) -> jnp.ndarray:
            """Generates random subkey"""
            self._key, subkey = jax.random.split(self._key)
            return subkey

    __instance = __random()

    @staticmethod
    def set_key(seed=None) -> None:
        """set_key"""
        random.__instance.set_key(seed)

    @staticmethod
    def get_key() -> jnp.ndarray:
        """get_key"""
        return random.__instance.get_key()

    @staticmethod
    def generate_key() -> jnp.ndarray:
        """generate_key"""
        return random.__instance.generate_key()



"""timecast.utils.gram"""
import jax
import jax.numpy as jnp

from timecast.utils.statistics import OnlineStatistics


class OnlineGram:
    """Compute gram matrix of observations online

    Notes:
        * Computes X.T @ X, where axis 0 is time and axis 1 is
        features
        * In the case where we supply a second matrix, this is no
        longer a gram matrix; X.T @ Y
    """

    def __init__(
        self,
        input_dim: int = 1,
        output_dim: int = None,
        fit_intercept: bool = True,
        normalize: bool = True,
    ) -> None:
        """Initialize OnlineGram

        Args:
            input_dim (int): feature dimension (typically X.shape[1])
            output_dim(int): same as input_dim unless we supply a
            second matrix with different dimensions
            fit_intercept (bool): fit intercept when finalized?
            normalize (bool): normalize when finalized?
        """
        self._input_dim = input_dim
        self._output_dim = output_dim or input_dim
        self._fit_intercept = fit_intercept
        self._normalize = normalize

        self._matrix = jnp.zeros((self._input_dim, self._output_dim))
        self._input_stats = OnlineStatistics(dim=self._input_dim)
        self._output_stats = None if output_dim is None else OnlineStatistics(dim=self._output_dim)

    @property
    def matrix(self):
        """Property to retrieve gram matrix"""
        result = self._matrix
        if self._normalize:
            result = self.normalize(matrix=result)

        if self._fit_intercept:
            result = self.fit_intercept(matrix=result)
        return result

    @property
    def observations(self):
        """Observations"""
        return self._input_stats.observations

    @property
    def mean(self):
        """Mean"""
        return self._input_stats.mean

    @property
    def std(self):
        """Std"""
        return self._input_stats.std

    def update(self, X: jnp.ndarray, Y: jnp.ndarray = None) -> None:
        """Update Gram matrix bassed on new data"""
        if Y is None and self._output_stats is not None:
            raise ValueError(
                "Expected second matrix since output_dim of {} was initialized".format(
                    self._output_dim
                )
            )
        if Y is not None and self._output_stats is None:
            raise ValueError("Unexpected second matrix since output_dim is None")

        Y = X if Y is None else Y

        if X.shape[0] != Y.shape[0]:
            raise ValueError(
                "Inputs must have the same number of observations; got {}, {}".format(
                    X.shape[0], Y.shape[0]
                )
            )
        self._matrix += X.T @ Y
        self._input_stats.update(X)
        if self._output_stats is not None:
            self._output_stats.update(Y)

    def normalize(self, matrix: jnp.ndarray = None) -> jnp.ndarray:
        """Z-score across feature dimensions"""
        X_stats = self._input_stats
        Y_stats = self._output_stats or self._input_stats
        matrix = self._matrix if matrix is None else matrix

        n = X_stats.observations

        result = (
            matrix
            + n * jnp.outer(X_stats.mean, Y_stats.mean)
            - jnp.outer(X_stats.mean, Y_stats.sum)
            - jnp.outer(X_stats.sum, Y_stats.mean)
        ) / jnp.outer(X_stats.std, Y_stats.std)

        # Replace nans with 0s because std of some columns may be 0
        result = jnp.nan_to_num(result)

        return result

    def fit_intercept(self, matrix: jnp.ndarray = None) -> jnp.ndarray:
        """Add an intercept to X for X.T @ Y

        Notes:
            * This method returns rather than modifies _matrix because
            the change in dimension would mean we can no longer update
        """
        matrix = self._matrix if matrix is None else matrix

        X_stats = self._input_stats
        Y_stats = self._output_stats
        result = jnp.concatenate((jnp.zeros((1, matrix.shape[1])), matrix), axis=0)

        if Y_stats is None:
            result = jnp.concatenate((jnp.zeros((result.shape[0], 1)), result), axis=1)

        stats = X_stats if Y_stats is None else Y_stats
        sumcol = stats.sum

        if self._normalize:
            sumcol = jnp.zeros_like(sumcol)

        if Y_stats is None:
            result = jax.ops.index_update(result, jax.ops.index[0, 0], stats.observations)
            result = jax.ops.index_update(result, jax.ops.index[:1, 1:], sumcol)
            result = jax.ops.index_update(result, jax.ops.index[1:, :1], sumcol.T)
        else:
            result = jax.ops.index_update(result, jax.ops.index[:1, :], sumcol)

        return result



"""Helper functions and classes for timecast.learners"""
import abc
import inspect
from typing import Iterable
from typing import Tuple

import flax
import jax.numpy as jnp
import numpy as onp

from timecast.utils import random


class FitMixin(abc.ABC):
    """Mixin class that provides a fit function for offline training / learner
    initialization"""

    @classmethod
    @abc.abstractmethod
    def fit(
        cls,
        data: Iterable[Tuple[onp.ndarray, onp.ndarray]],
        input_dim: int,
        output_dim: int = 1,
        key: jnp.ndarray = None,
        **kwargs
    ) -> flax.nn.Model:
        """Fit and initialize learner on training data

        Notes:
            * We could infer input_dim from data, but for now, require
            users to explicitly provide
            * output_dim defaults to 1 and is ignored for now

        Todo:
            * Really intended for passing in timeseries at a time, not
            individual time series observations; is this the right general API?
            * Shape is (1, input_dim); what about mini-batches?

        Args:
            data: an iterable of tuples containing input/truth pairs of time
            series
            input_dim: number of feature dimensions in input
            output_dim: number of feature dimensions in output
            key: random key for jax random
            kwargs: Extra keyword arguments

        Returns:
            flax.nn.Model: initialized model
        """
        raise NotImplementedError()


class NewMixin:
    """Mixin class that provides a flax.nn.Model constructor function"""

    # TODO (flax)
    @classmethod
    def new(
        cls, shape: Tuple[int, ...], *, key: jnp.ndarray = None, name: str = None, **kwargs
    ) -> flax.nn.Model:
        """
        Args:
            shape (Tuple[int, ...]): shape for initialization
            key (jnp.ndarray): key for jax random
            name (str): identifier for top-level module
            **kwargs: arguments for flax.nn.Module
        """
        argspec = inspect.getfullargspec(cls.apply)

        num_kwargs = 0 if argspec.defaults is None else len(argspec.defaults)

        _kwargs = {}

        # Capture kwarg defaults
        for i in range(1, num_kwargs + 1):
            _kwargs[argspec.args[-i]] = argspec.defaults[-i]

        # Override kwarg defaults, if any
        for kw, arg in kwargs.items():
            if kw in argspec.args or argspec.varkw is not None:
                _kwargs[kw] = arg
            else:
                raise ValueError("Found extra kwarg {}: {}".format(kw, arg))

        model_def = cls.partial(**kwargs)

        if name is None:
            name = cls.__name__.lower()

        if key is None:
            key = random.generate_key()

        with flax.nn.stateful() as state:
            _, params = model_def.init_by_shape(key, [shape])
        model = flax.nn.Model(model_def, params)

        return model, state



"""timecast.learners"""
from timecast.learners._ar import AR
from timecast.learners._combinators import Parallel
from timecast.learners._combinators import Sequential
from timecast.learners._index import Index
from timecast.learners._pcr import PCR
from timecast.learners._predict_constant import PredictConstant
from timecast.learners._predict_last import PredictLast
from timecast.learners.base import FitMixin


__all__ = [
    "AR",
    "Index",
    "FitMixin",
    "PCR",
    "PredictConstant",
    "PredictLast",
    "Parallel",
    "Sequential",
]



"""timecast.learners.PredictLast: testing"""
import flax
import jax
import numpy as onp
import pytest

from timecast.learners import PredictLast
from timecast.utils import random

shapes = [(4, 32), (10,), (10, 1), (1,), (1, 10)]


def create_predict_last(shape):
    """Generate PredictLast model and state"""
    with flax.nn.stateful() as state:
        _, params = PredictLast.init_by_shape(jax.random.PRNGKey(0), [shape])
        model = flax.nn.Model(PredictLast, params)
    return model, state


@pytest.mark.parametrize("shape", shapes)
def test_predict_last(shape):
    """Test PredictLast"""
    model, state = create_predict_last(shape)

    X = jax.random.uniform(random.generate_key(), shape=shape)
    with flax.nn.stateful(state) as state:
        ys = model(X)

    onp.testing.assert_array_almost_equal(X, ys)



"""timecast.series.lstm"""
from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as onp


def generate(
    n: int = 1000, input_dim: int = 5, output_dim: int = 1, hidden_dim: int = 64
) -> Tuple[onp.ndarray, onp.ndarray]:
    """
    Description: output from a randomly initialized LSTM
    """
    W_hh = jnp.asarray(onp.random.rand(4 * hidden_dim, hidden_dim))
    W_xh = jnp.asarray(onp.random.rand(4 * hidden_dim, input_dim))
    b_h = onp.zeros(4 * hidden_dim)
    b_h[hidden_dim : 2 * hidden_dim] = onp.ones(hidden_dim)
    b_h = jnp.asarray(b_h)
    W_out = jnp.asarray(onp.random.rand(output_dim, hidden_dim))
    cell = jnp.zeros(hidden_dim)
    hidden = jnp.zeros(hidden_dim)

    def sigmoid(x):
        """Sigmoid function"""
        return 1.0 / (1.0 + jnp.exp(-x))

    def step(hc, x):
        """Internal function for LSTM step"""
        h, c = hc
        gate = jnp.dot(W_hh, h) + jnp.dot(W_xh, x) + b_h
        i, f, g, o = onp.split(gate, 4)
        nc = sigmoid(f) * c + sigmoid(i) * jnp.tanh(g)
        nh = sigmoid(o) * jnp.tanh(nc)
        y = jnp.dot(W_out, nh)
        return (nh, nc), y

    X = jnp.asarray(onp.random.rand(n, input_dim))
    _, y = jax.lax.scan(step, (hidden, cell), X)

    return X, y.reshape(-1, 1)



"""timecast: an library for online time series analysis"""
import datetime
import os
import sys

from setuptools import find_packages
from setuptools import setup

name = "timecast"
version = "0.2.7"

if "--nightly" in sys.argv:
    name += "-nightly"

    date_suffix = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    version += ".dev" + date_suffix
    sys.argv.remove("--nightly")
    print(name, version)

with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md"), encoding="utf-8",
) as f:
    long_description = f.read()

setup(
    name=name,
    version=version,
    author="Google AI Princeton",
    author_email="dsuo@google.com",
    description="Performant, composable online learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MinRegret/timecast",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=["timeseries", "time", "series", "analysis", "online", "learning"],
    python_requires=">=3.6",
    install_requires=["numpy", "jax", "jaxlib", "flax", "pandas", "pathos", "tqdm"],
    extras_require={
        "dev": [
            "flake8",
            "flake8-print",
            "flake8-bugbear",
            "mypy",
            "scikit-learn",
            "pytest",
            "pytest-xdist",
            "pytest-cov",
            "pylint",
            "black",
            "reorder-python-imports",
            "autoflake",
            "pre-commit",
            "pydocstring-coverage",
            "bumpversion",
            "ipython",
            "jupyter",
            "gprof2dot",
            "bandit",
            "check-python-versions",
            "line-profiler",
        ],
        "docs": ["sphinx", "sphinxcontrib-napoleon", "sphinx_rtd_theme"],
    },
)



"""timecast.learners.base: testing"""
import flax
import pytest

from timecast.learners.base import FitMixin
from timecast.learners.base import NewMixin


def test_fit_mixin_abc():
    """Test that FitMixin is abstract"""
    with pytest.raises(TypeError):
        FitMixin()


def test_fit_mixin_subclass():
    """Test that subclass must implement abstract classes"""

    class Dummy(FitMixin):
        """Dummy class"""

    with pytest.raises(TypeError):
        Dummy()


def test_fit_mixin_fit():
    """Test unimplemented class method"""
    with pytest.raises(NotImplementedError):
        FitMixin.fit([], 1)


def test_new_mixin():
    """Test common case"""

    class DummyNew(NewMixin, flax.nn.Module):
        """Dummy class"""

        def apply(self, x, a=2, b=3):
            """Dummy apply"""
            return x + 2 + 3

    DummyNew.new((1, 2))


def test_new_mixin_multiple_args():
    """Test multiple unnamed args"""

    class DummyNew(NewMixin, flax.nn.Module):
        """Dummy class"""

        def apply(self, x, a, b=3):
            """Dummy apply"""
            return x + 2 + 3

    with pytest.raises(TypeError):
        DummyNew.new((1, 2))

    DummyNew.new((1, 2), a=2)

    with pytest.raises(ValueError):
        DummyNew.new((1, 2), c=4)



"""timecast.learners.pcr: testing"""
import jax
import numpy as onp
import pytest
from sklearn.decomposition import PCA

from timecast.learners._pcr import _compute_pca_projection
from timecast.utils import random


@pytest.mark.parametrize("shape", [(1, 1), (10, 1), (2, 10), (10, 2)])
def test_compute_pca_projection(shape):
    """Test PCA projection of X vs X.T @ X"""
    X = jax.random.uniform(random.generate_key(), shape=shape)
    XTX = X.T @ X

    k = 1 if X.ndim == 1 else min(X.shape)
    p1 = _compute_pca_projection(X, k)
    p2 = _compute_pca_projection(XTX, k)

    onp.testing.assert_array_almost_equal(abs(p1), abs(p2), decimal=3)


@pytest.mark.parametrize("shape", [(1, 1), (10, 1), (1, 10), (10, 10)])
def test_compute_pca_projection_sklearn(shape):
    """Test PCA projection of X vs sklearn"""
    X = jax.random.uniform(random.generate_key(), shape=shape)

    projection = _compute_pca_projection(X, 1, center=True)

    pca = PCA(n_components=1)
    pca.fit(X)

    onp.testing.assert_array_almost_equal(abs(projection), abs(pca.components_.T), decimal=3)



"""timecast.series.lds"""
from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as onp


def generate(
    n: int = 1000, input_dim: int = 5, output_dim: int = 1, hidden_dim: int = 5, noise: float = 1.0
) -> Tuple[onp.ndarray, onp.ndarray]:
    """
    Description: Simulates a randomly initialized linear dynamical system
    """

    def normalize(M, k):
        """Shrinks matrix M such that largest eigenvalue has magnitude k"""
        return k * M / jnp.linalg.norm(M, ord=2)

    A = jnp.asarray(onp.random.rand(hidden_dim, hidden_dim))
    B = jnp.asarray(onp.random.rand(hidden_dim, input_dim))
    C = jnp.asarray(onp.random.rand(output_dim, hidden_dim))
    D = jnp.asarray(onp.random.rand(output_dim, input_dim))
    h = jnp.asarray(onp.random.rand(hidden_dim))

    A = normalize(A, 1.0)
    B = normalize(B, 1.0)
    C = normalize(C, 1.0)
    D = normalize(D, 1.0)

    def step(h, x):
        """Internal function for LDS step"""
        x, eps_h, eps_y = x
        h = jnp.dot(A, h) + jnp.dot(B, x) + noise * eps_h
        y = jnp.dot(C, h) + jnp.dot(D, x) + noise * eps_y
        return h, y

    X = jnp.asarray(onp.random.rand(n, input_dim))
    EPS_H = jnp.asarray(onp.random.rand(n, hidden_dim))
    EPS_Y = jnp.asarray(onp.random.rand(n, output_dim))

    _, y = jax.lax.scan(step, h, (X, EPS_H, EPS_Y))

    return X, y.reshape(-1, 1)



"""timecast.utils.internalize: testing"""
import jax.numpy as jnp
import numpy as onp
import pytest

from timecast.utils import internalize


@pytest.mark.parametrize(
    "n,shape,expected",
    [
        (1, (1,), (True, 1)),
        (1, (10,), (False, jnp.ones((10, 1)))),
        (1, (1, 1), (True, 1)),
        (1, (10, 1), (False, jnp.ones((10, 1)))),
        (2, (2,), (True, jnp.ones((1, 2)))),
        (2, (1, 2), (True, jnp.ones((1, 2)))),
        (2, (4, 2), (False, jnp.ones((4, 2)))),
        (1, (), (True, 1)),
    ],
)
def test_internalize(n, shape, expected):
    """Test internalize"""
    X = jnp.ones(shape)

    X, is_value, dim, _ = internalize(X, n)

    assert is_value == expected[0]
    onp.testing.assert_array_equal(X, expected[1])


@pytest.mark.parametrize(
    "n,shape",
    [(2, (1, 3)), (2, (3, 1)), (2, (2, 4)), (2, (2, 1)), (2, (5,)), (1, (2, 2)), (1, (1, 10))],
)
def test_internalize_value_error(n, shape):
    """Test value error"""
    with pytest.raises(ValueError):
        internalize(jnp.ones(shape), n)


def test_internalize_type_error():
    """Test type error"""
    with pytest.raises(TypeError):
        internalize(jnp.ones((1, 2, 3)), 10)


@pytest.mark.parametrize("n,expected", [(1, True), (10, False)])
def test_internalize_scalar(n, expected):
    """Test scalar"""
    if expected:
        _, is_value, _, _ = internalize(4, n)
        assert is_value == expected
    else:
        with pytest.raises(ValueError):
            internalize(4, n)



"timecast.series._core"
import jax.numpy as jnp
import pandas as pd


def generate_timeline(path: str, name=None, delimiter=","):
    """Convenience function to grab a single time series and convert to X, y pair"""
    data = pd.read_csv(path, delimiter=delimiter)

    if name is not None:
        data = data[[name]]

    return jnp.asarray(data)[:-1].reshape(-1, 1), jnp.asarray(data)[1:].reshape(-1, 1)



"""timecast.utils.statistics"""
from numbers import Real
from typing import Union

import jax.numpy as jnp

from timecast.utils import internalize


class OnlineStatistics:
    """Compute summary statistics online"""

    def __init__(self, dim: int = 1):
        """Initialize OnlineStatistics"""
        self._dim = dim
        self._mean = jnp.zeros((1, dim))
        self._var = jnp.zeros((1, dim))
        self._sum = jnp.zeros((1, dim))
        self._observations = 0

    def update(self, observation: Union[Real, jnp.ndarray]) -> None:
        """Update with new observation"""
        observation, is_value, _, _ = internalize(observation, self._dim)

        num_observations = observation.shape[0]

        prev_mean = self._mean
        curr_mean = observation if is_value else observation.mean(axis=0)
        self._mean = (self._observations * prev_mean + num_observations * curr_mean) / (
            self._observations + num_observations
        )

        prev_var = self._var
        curr_var = 0 if is_value else observation.var(axis=0)

        self._var = (
            self._observations * prev_var
            + num_observations * curr_var
            + self._observations * ((prev_mean - self._mean) ** 2)
            + num_observations * ((curr_mean - self._mean) ** 2)
        ) / (self._observations + num_observations)

        self._sum += observation if is_value else observation.sum(axis=0)

        self._observations += num_observations

    @property
    def mean(self) -> Union[Real, jnp.ndarray]:
        """Mean"""
        return self._mean

    @property
    def var(self) -> Union[Real, jnp.ndarray]:
        """Variance"""
        return self._var

    @property
    def std(self) -> Union[Real, jnp.ndarray]:
        """Standard deviation"""
        return jnp.sqrt(self._var)

    @property
    def sum(self) -> Union[Real, jnp.ndarray]:
        """Sum"""
        return self._sum

    @property
    def observations(self) -> int:
        """Number of observations"""
        return self._observations

    def zscore(self, data: jnp.ndarray) -> Union[Real, jnp.ndarray]:
        """Zscore an observation based on current statistics"""
        return (data - self.mean) / self.std
